/*  Multiple precision integer arithmetic - see multip1.notes	*/
/*  This file has addition, subtraction, multiplication, and	*/
/*  utility routines.  See midiv.c  for division.		*/
/*  Source:  miarith.c		Header:  miarith.h		*/
/*  Ernest Sibert -- May 1995					*/

#include <stdio.h>
#include "multip.h"

/* -------------------------------------------------------------------- */

/*  Integer  d := x + y, may overflow	*/
/*  d == x or d == y is permitted.	*/

void imadd(int d[], int x[], int y[], int N)
 { int i, cy, w, nz, xs, ys;
   xs = x[0];  ys = y[0];		/* Status words */
   cy = nz = 0;				/* Carry value, non-zero detector */
   if ( (xs|ys)&Ovflw ) 		/* Overflow in input(s)	*/
      d[0] = Ovflw;
   else					/* A real addition	*/
    { for ( i = N;  i > 0;  i-- )
       { w  = x[i] + y[i] + cy;
	 cy = ( w&Bcy ) ? 1 : 0;  w &= Bwrd;
	 d[i] = w;  nz |= w;
       }
      /* Result status - Nonzr Ovflw Sign */
      w = (((xs&Sign) + (ys&Sign) + cy)^(cy<<1)) | (nz > 0)<<2;
      if ( w&Sign ) w |= Bneg;		/* Fill in neg. sign */
      d[0] = w;
    }
 }

/* -------------------------------------------------------------------- */

/*  Integer  d := -n, may overflow	*/
/*  d == x is permitted.		*/

void imneg(int d[], int x[], int N)
 { int i, cy, w, nz, xs;
   xs = x[0];  cy = 1;  nz = 0;
   if ( xs&Ovflw )  d[0] = Ovflw;
   else
    { for ( i = N;  i > 0;  i-- )
       { w  = ((~x[i])&Bwrd) + cy;
	 cy = ( w&Bcy ) ? 1 : 0;  w &= Bwrd;
	 d[i] = w;  nz |= w;
       }
      /* Result status - Nonzr Ovflw Sign */
      w = ((((~xs)&Sign) + cy)^(cy<<1)) | (nz > 0)<<2;
      if ( w&1 ) w |= Bneg;		/* Fill in neg. sign */
      d[0] = w;
    }
 }

/* -------------------------------------------------------------------- */

/*  Integer  d := x - y, may overflow	*/
/*  d == x or d == y is permitted.	*/

void imsub(int d[], int x[], int y[], int N)
 { int i, cy, w, nz, xs, ys;
   xs = x[0];  ys = y[0];		/* Status words */
   cy = 1;  nz = 0;			/* Carry value, non-zero detector */
   if ( (xs|ys)&Ovflw ) d[0] = Ovflw;	/* Overflow in input(s)	*/
   else					/* A real subtraction	*/
    { for ( i = N;  i > 0;  i-- )
       { w  = x[i] + ((~y[i])&Bwrd) + cy;
	 cy = ( w&Bcy ) ? 1 : 0;  w &= Bwrd;
	 d[i] = w;  nz |= w;
       }
      /* Result status - Nonzr Ovflw Sign */
      w = (((xs&Sign) + ((~ys)&Sign) + cy)^(cy<<1)) | (nz > 0)<<2;
      if ( w&1 ) w |= Bneg;		/* Fill in neg. sign */
      d[0] = w;
    }
 }

/* -------------------------------------------------------------------- */

/*  d := x for multiple-precision	*/
/*  d == x is permitted but silly..	*/

void  mcopy(int d[], int x[], int N)
 { int i;
   for ( i = 0;  i <= N;  i++ )  d[i] = x[i];
 }

/* -------------------------------------------------------------------- */

/*  d := 0 for multiple-precision	*/

void  mzero(int d[], int N)
 { int i;
   for ( i = 0;  i <= N;  i++ )  d[i] = 0;
 }

/* -------------------------------------------------------------------- */

/*  Integer  dd[0..M+N] := x[0..M]*y[0..N], for	*/
/*  x >= 0 and y >= 0, whence result is >= 0.	*/
/*  Neither x nor y should interfere with dd.	*/
/*  Overflow is impossible.			*/
/*  M = 1, N > 1 is better than M > 1, N = 1 if	*/
/*  either case is likely.			*/

void imult2x(int dd[], int x[], int y[], int M, int N)
 { int i, j, k, cy, w, nz, xs, ys;
   int xlo, xhi, ylo, yhi, w1, w0, p2, p1, p0;
   xs = x[0];  ys = y[0];
   mzero(dd, M+N);
   cy = 0;			/* Carry from earlier steps	*/
   /*  Diagnostic printing ********************************************
       printf(
    "i j k      p2         p1         p0         w1         w0     cy\n");
       End diagnostic printing ************************************** */
   for ( i = M;  i > 0; i-- )
    { xlo = x[i];  xhi = (xlo&Bhi)>>Mhalf;  xlo &= Blo;
      w0 = dd[i+N];		/* To start loop */
      for ( j = N;  j > 0;  j--)
       { k = i + j;		/* Position to add items to dd	*/
	 /* Now w0 = effective dd[k]; */
	 ylo = y[j];  yhi = (ylo&Bhi)>>Mhalf;  ylo &= Blo;
	 p0  = xlo*ylo;			/* Leaves p0 < 2^30		*/
	 p1  = xlo*yhi + xhi*ylo;	/* May yield carry, but OK	*/
	 p2  = xhi*yhi + cy;		/* Absorbs cy, leaves p2 < 2^30	*/
	 w0 += p0;			/* May yield carry	*/
	 cy  = (w0&Bcy) ? 1 : 0;  w0 &= Bwrd;
	 w0 += (p1&Blo)<<Mhalf;		/* May yield carry	*/
	 cy += (w0&Bcy) ? 1 : 0;  w0 &= Bwrd;
	 w1  = dd[k-1];
	 w1 += (p1>>15) + p2 + cy;	/* May yield carry	*/
	 cy  = (w1&Bcy) ? 1 : 0;  w1 &= Bwrd;
	 dd[k] = w0;			/* dd[k] is done until next i	*/
    /*  Diagnostic printing ********************************************
	printf("%d %d %d  0x%08x 0x%08x 0x%08x 0x%08x 0x%08x %2d\n",
		   i, j, k,   p2,    p1,    p0,    w1,    w0,   cy    );
        End diagnostic printing ************************************** */
	 w0 = w1;			/* for next j, no need to store	*/
       }	/* end for ( j ) */
      dd[i] = w1;
    }     /* end for ( i ) */
   dd[0] = xs&ys&Nonzr;		/* Status for non-neg product */
 }

  /**************************************************************/
  /* When multiplying x[i]*y[j],  k = i+j:			*/
  /*								*/
  /*   31   29     15 14      0     31   29     15 14      0	*/
  /*                      cy(in)				*/
  /*  +------------------------+   +------------------------+   */
  /*  |   0|      dd[k-1]      |   |   0|   dd[k] = w0      |	*/
  /*  +------------------------+   +------------------------+   */
  /*                               +------------------------+   */
  /*                               |   0|   p0    :         |	*/
  /*                               +------------------------+   */
  /*            +---------------........----------+       	*/
  /*            |   c|         :  p1    :         |		*/
  /*            +---------------........----------+		*/
  /*  +------------------------+				*/
  /*  |   0|    p2   :         |				*/
  /*  +------------------------+				*/
  /*  cy(out)							*/
  /*								*/
  /*  Note that cy(in) = 0 when j = N				*/
  /*       and cy(out) = 0 when	j = 1				*/
  /**************************************************************/

/* -------------------------------------------------------------------- */

/*  Integer  dd[0..M+N] := x[0..M]*y[0..N]			*/
/*  Multiplication by the most negative integer yelds overflow,	*/
/*  all other cases succeed without overflow.			*/
/*  This program buffers x and y, hence is not sensitive to	*/
/*  interference.						*/

void imult2(int dd[], int x[], int y[], int M, int N)
 { int xs, ys;
   int *xr,  *yr;		/* buffers for absolute values	*/
   xr = (int *) malloc(Hbyte*(M+1));
   yr = (int *) malloc(Hbyte*(N+1));
   if ( (xr == NULL) | (yr == NULL) )			/* Failure	*/
    { fprintf(stderr, "Out of memory in imult2.\n");
      dd[0] = Ovflw;  return;
    }
   xs = x[0];  ys = y[0];		/* Status words		*/
   /*  Form absolute values in xr, yr	*/
   if ( xs&Sign ) imneg(xr, x, M); else mcopy(xr, x, M);
   if ( ys&Sign ) imneg(yr, y, N); else mcopy(yr, y, N);
   if ( (xs | ys | xr[0] | yr[0]) & Ovflw ) dd[0] = Ovflw;
   else
    { imult2x(dd, xr, yr, M, N);
      if ( (xs^ys)&Sign ) imneg(dd, dd, M+N);
    }     /* end else */
   free(xr);  free(yr);			/* Return buffer space */
 }

/* -------------------------------------------------------------------- */

/*  x[0..N] := s*x[0..N] + d  where  x >= 0 and  0 <= d < s < 2^Mhalf .	*/
/*  May overflow							*/

void immuls(int x[], int N, int s, int d)
 { int i, xlo, xhi , cy, nzr;
   cy = d;  nzr = 0;
   for ( i = N;  i >= 1;  i-- )
    { xlo = x[i];  xhi = (xlo&Bhi)>>Mhalf;  xlo &= Blo;
      xlo = s*xlo + cy;
       cy = (xlo&Bhi)>>Mhalf;  xlo &= Blo;
      xhi = s*xhi + cy;
       cy = (xhi&Bhi)>>Mhalf;  xhi &= Blo;
      xhi = (xhi<<Mhalf) | xlo;
      nzr |= xhi;  x[i] = xhi;
    }
   x[0] |= ((nzr > 0)<<2) | ((cy > 0)<<1);
 } 

/* -------------------------------------------------------------------- */

/*  Shift, d := x*2^e, may generate overflow.	*/
/*  d[0..N] should not interfere with x[0..N].	*/

void imshift(int d[], int x[], int e, int N)
 { int swrds, sbits, ae, sbdwn, i, j, fill, mask, u, w, nzr, ds;
   ae = abs(e);  swrds = ae/Mword;  sbits = ae%Mword;  sbdwn = Mword - sbits;
   if ( swrds > N ) swrds = N;
   ds = x[0];			/* Status */
   fill = ( d[0]&Sign ) ? Bwrd : 0;
   if ( e <= 0 )	/* Shift right */
    { for ( i = 1; i <= swrds; i++ ) d[i] = fill;
      nzr = ( 1 <= swrds ) ? fill : 0;	/* To check for non-zero bits	 */
      mask = Bwrd>>sbdwn;		/* Bits which cross word	 */
      fill = (fill&mask)<<sbdwn;	/* For first word with data	 */
      for ( j = 1 ; i <= N; i++, j++ )	/* Fill words which receive data */
       { w    = x[j];
	 d[i] = u = (w>>sbits) | fill;  nzr |= u;
	 fill = (w&mask)<<sbdwn;		/* To next word	*/
       }
    }	/* end of case e <= 0 */
   else	/* shift left */
    { u = N - swrds;		/* To fill 0 words at right */
      for ( i = N;  i > u;  i-- ) d[i] = 0;
      mask = Bwrd>>sbits;		/* Bits which stay in word	 */
      nzr = 0;  u = 0;
      for ( j = N;  i >= 1;  i--, j-- )	/* Fill words which receive data */
       { w    = x[j];
         d[i] = u = ((w&mask)<<sbits) | u;  nzr |= u;
	 u    = w>>sbdwn;			/* To next word */
       }
      u ^= (fill>>sbdwn);		/* To check for overflow */
      for (  ;  j >= 1;  j-- ) u |= (x[j]^fill);
      if ( u ) ds |= Ovflw;
    }
   if ( nzr ) ds |= Nonzr; else ds &= ~Nonzr;
   d[0] = ds;
 }

/* -------------------------------------------------------------------- */

/*  Report status of mp quantity as:		*/
/*       4   overflow				*/
/*	 1   positive				*/
/*	 0   zero				*/
/*	-1   negative, but not most negative	*/
/*	-2   most negative integer 		*/

int imstat(int x[])
 { int xs;
   xs = x[0];
   return ( xs&Ovflw ) ? 4 : xs>>2;
 }

/* -------------------------------------------------------------------- */

/*  Compare mp integers x, y, result is:	*/
/*	 4   overflow ( x or y has overflow)	*/
/*	 1   x > y				*/
/*       0   x = y				*/
/*	-1   x < y				*/

int imcomp(int x[], int y[], int N)
 { int i, xs, ys;
   xs = x[0];  ys = y[0];
   if ( (xs|ys)&Ovflw) return 4;
   for ( i = 0;  i <= N;  i++ )
    if ( x[i] > y[i] ) return  1; else
    if ( x[i] < y[i] ) return -1;
   return 0;
 }

/* -------------------------------------------------------------------- */

/*  Determine position of leading 1 in x, relative to mp words	*/
/*  Positions are counted from top.				*/
/*  Example for 32-bit machine word, 30 bits mp:		*/
/*								*/
/*            0   1   2   3   4   5   6   7   8			*/
/*  +---+---+---+---+---+---+---+---+---+---+---+ . . .		*/
/*  | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 0 | 1 | 1 |		*/
/*  +---+---+---+---+---+---+---+---+---+---+---+ . . .		*/
/*    31  30  29  28  27  26  25  24  23  22  21		*/
/*								*/
/*  The result is 5, which is the left-shift to normalize.	*/

int Bitpos(int x)
 { int p;
   p = 0;
   if ( (x&Bc16) == 0 ) p += 16; else x &= Bc16;
   if ( (x&Bc08) == 0 ) p +=  8; else x &= Bc08;
   if ( (x&Bc04) == 0 ) p +=  4; else x &= Bc04;
   if ( (x&Bc02) == 0 ) p +=  2; else x &= Bc02;
   if ( (x&Bc01) == 0 ) p +=  1; else x &= Bc01;
   return ( p - (Hword - Mword) );
 }

/* -------------------------------------------------------------------- */

/*  Normalize x > 0 so that x[1] >= (1/2)*s^Wsize [ 2^29 ]	*/
/*  Returns number of bits required for shift.  Returns -1 if	*/
/*  x <= 0.							*/

int imnorm(int x[], int N)
 { int xs, i, j, k, s, sx, fill, w;
   xs = x[0];					/* Status	*/
   if ( (xs&Nonzr == 0) | (xs&Sign) ) return -1;
   else
    { for ( i = 1;  x[i] == 0;  i++ ) ;		/* Leading x[i] > 0	*/
      if ( i > 1 )				/* Shift words		*/
       { for ( j = 1, k = i;  k <= N;  j++, k++ ) x[j] = x[k];
	 for ( k = j;  k <= N;  k++ ) x[k] = 0; /* Zero remaining words	*/
       }
      else j = N+1;				/* All words will shift	*/
      s = Bitpos(x[1]);			/* Bits to shift	*/
      if ( s > 0 )				/* Shift bits		*/
       { fill = 0;  sx = Mword - s;
	 for ( k = j-1;  k >= 1;  k-- )
	  { w    = x[k];
	    x[k] = (w << s) | fill;
	    fill = w >> sx;
	  }
       }
      return ((i-1)*Mword + s);	/* Shift distance (bits)	*/
    }
 }
